/* PNAT.  Simple Port Trigger */
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/timer.h>
#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/netdevice.h>
#include <linux/if.h>
#include <linux/skbuff.h>
#include <linux/udp.h>
#include <linux/inetdevice.h>
#include <net/protocol.h>
#include <net/checksum.h>
#include <linux/proc_fs.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>
#include "ipt_PNAT.h"

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

struct ipt_pnat_info pnat_info[MAX_PORT_TRIGGER];

static unsigned long get_expire(const struct ip_conntrack *conntrack)
{
	if(!conntrack)
		return 0;
	printk("get_expire\n");
	return (timer_pending(&conntrack->timeout) ? (conntrack->timeout.expires - jiffies)/HZ : 0);
}
#if 0
static unsigned int print_conntrack(char *buffer, const struct ip_conntrack *conntrack)
{
	unsigned int len;
	if(!conntrack)
		return 0;
	len = sprintf(buffer, "%u %lu\n",
		      conntrack->tuplehash[IP_CT_DIR_ORIGINAL].tuple.dst.protonum
		      ,timer_pending(&conntrack->timeout) ? (conntrack->timeout.expires - jiffies)/HZ : 0);
	return len;
}
#endif

/* add ct to list */
void add_ct_list(struct ipt_pnat_info *info,struct ip_conntrack *ct)
{
	struct ip_conntrack_list *list=NULL;
	struct ip_conntrack_list *last=NULL;
	list=info->ct_list;
	while(list){
		last=list;
		list=list->next;
	}
	if(last==NULL){
		info->ct_list=kmalloc(sizeof(struct ip_conntrack_list), GFP_ATOMIC);	
		info->ct_list->ct=ct;
		info->ct_list->next=NULL;
	}else{
		last->next=kmalloc(sizeof(struct ip_conntrack_list), GFP_ATOMIC);
		last->next->ct=ct;
		last->next->next=NULL;
	}
}

/* check is timeout or not */
static unsigned long check_expire(struct ipt_pnat_info *info)
{
	struct ip_conntrack_list *prev=NULL;
	struct ip_conntrack_list *list=NULL;
	unsigned long expire_time=0;
	unsigned long max_expire_time=0;
	list=info->ct_list;
	printk("enter check_expire\n");
	while(list){
		expire_time=get_expire(list->ct);
		
		if(max_expire_time < expire_time)
			max_expire_time = expire_time;
#if 0		
		if(expire_time==0){
			list=list->next;
			/* not frist entry */
			if(prev){
				kfree(prev->next);
				prev->next=list;
			}
		}
		prev=list;
#endif
		list=list->next;
	}
	printk("max_expire_time=%ld\n",max_expire_time);
	return max_expire_time;
}

/* free ct_list */
void free_ct_list(struct ipt_pnat_info *info)
{
	struct ip_conntrack_list *list=NULL;
	struct ip_conntrack_list *next=NULL;
	
	list=info->ct_list;
	while(list){
		next=list->next;
		kfree(list);
		list=next;
	}
}
/* find outbound ifname proto and port range */
struct ipt_pnat_info* find_outbound(char *ifname,u_int16_t protocol,u_int16_t port)
{
	int i=0;
//	printk("find_outbound ifname==%s protocol==%d port==%u\n",ifname,protocol,port);
	for(i=0;i<MAX_PORT_TRIGGER;i++){
		if(pnat_info[i].outbound.ifname[0]==0)	
			continue;
		/* not use */
		if(pnat_info[i].outbound.min==0)
			continue;

		if(strcmp(pnat_info[i].outbound.ifname,ifname)!=0)
			continue;

	//	printk("out port=%u %u %u\n",port,pnat_info[i].outbound.min,pnat_info[i].outbound.max);			
		/* tcp or both*/
		if((protocol==IPPROTO_TCP && strcmp(pnat_info[i].outbound.protocol,"udp")!=0)
			/* udp or both */
			||( protocol==IPPROTO_UDP && strcmp(pnat_info[i].outbound.protocol,"tcp")!=0)){
			
				if(port >= pnat_info[i].outbound.min 
						&& port <= pnat_info[i].outbound.max)
					return &pnat_info[i];
		}
	}
	return NULL;
}
/* find inbound ifname proto and port range */
struct ipt_pnat_info* find_inbound(char *ifname,u_int16_t protocol,u_int16_t port)
{
	int i=0;
	
//	printk("find_inbound ifname==%s protocol==%d port==%u\n",ifname,protocol,port);
	for(i=0;i<MAX_PORT_TRIGGER;i++){
		if(pnat_info[i].inbound.ifname[0]==0)	
			continue;
#if 0	
		/* trigger unknown */
		if(pnat_info[i].trigger_ip==0)
			continue;
#endif
		/* not use */
		if(pnat_info[i].inbound.min==0)
			continue;
		/* compare name */
		if(strcmp(pnat_info[i].inbound.ifname,ifname)!=0)
			continue;

	//	printk("in port=%u %u %u\n",port,pnat_info[i].inbound.min,pnat_info[i].inbound.max);			
		/* tcp or both*/
		if((protocol==IPPROTO_TCP && strcmp(pnat_info[i].inbound.protocol,"udp")!=0)
			/* udp or both */
			||( protocol==IPPROTO_UDP && strcmp(pnat_info[i].inbound.protocol,"tcp")!=0)){
			
				if(port >= pnat_info[i].inbound.min 
						&& port <= pnat_info[i].inbound.max)
					return &pnat_info[i];
		}
	}
	return NULL;
}

/* FIXME: Take multiple ranges --RR */
static int pnat_check(const char *tablename,
	       const struct ipt_entry *e,
	       void *targinfo,
	       unsigned int targinfosize,
	       unsigned int hook_mask)
{
	const struct ip_nat_multi_range *mr = targinfo;

	if (strcmp(tablename, "nat") != 0) {
		DEBUGP("pnat_check: bad table `%s'.\n", table);
		return 0;
	}
	if (targinfosize != IPT_ALIGN(sizeof(*mr))) {
		DEBUGP("pnat_check: size %u.\n", targinfosize);
		return 0;
	}
	if (hook_mask & ~((1 << NF_IP_PRE_ROUTING))) {
		DEBUGP("pnat_check: bad hooks %x.\n", hook_mask);
		return 0;
	}
	if (mr->range[0].flags & IP_NAT_RANGE_MAP_IPS) {
		DEBUGP("pnat_check: bad MAP_IPS.\n");
		return 0;
	}
	if (mr->rangesize != 1) {
		DEBUGP("pnat_check: bad rangesize %u.\n", mr->rangesize);
		return 0;
	}
	return 1;
}



static unsigned int pnat_target(struct sk_buff **pskb,
		unsigned int hooknum,
		const struct net_device *in,
		const struct net_device *out,
		const void *targinfo,
		void *userinfo)
{
	enum ip_conntrack_info ctinfo;
	//const struct ip_nat_multi_range *mr = targinfo;
	struct ip_nat_multi_range newrange;
	struct iphdr *iph = (*pskb)->nh.iph;
	void *protoh = (u_int32_t *)iph + iph->ihl;
	u_int16_t port;	
	struct ip_conntrack *ct;
	struct ipt_pnat_info *pnat;
	char name[16];
		
	ct = ip_conntrack_get(*pskb, &ctinfo);
	
	IP_NF_ASSERT(ct && (ctinfo == IP_CT_NEW || ctinfo == IP_CT_RELATED));
	
	
	IP_NF_ASSERT(hooknum == NF_IP_PRE_ROUTING || hooknum == NF_IP_LOCAL_OUT);
	
	if(hooknum==NF_IP_LOCAL_OUT)
		return IPT_CONTINUE;	

	if(!in->name)
		return IPT_CONTINUE;
	
	strcpy(name,in->name);
	
	switch (iph->protocol){
		case IPPROTO_TCP:
		{
			struct tcphdr *tcph = protoh;
			port=ntohs(tcph->dest);
			break;
		}
			
		case IPPROTO_UDP:
		{
			struct udphdr *udph = protoh;
			port=ntohs(udph->source);
			
			break;
		 }
			
		default:
			/* only support udp and tcp */
			return IPT_CONTINUE;
	}
	/* LAN to WAN is outbound */
	pnat=find_outbound(name,iph->protocol,port);
	if(pnat){
		/* time out or not use */
		if(pnat->trigger_ip==0 || check_expire(pnat)==0){
			pnat->trigger_ip=iph->saddr;
			free_ct_list(pnat);
		}
		printk("add_ct_list\n");
		add_ct_list(pnat,ct);
		
		return IPT_CONTINUE;	
	}
	
	/* WAN to LAN is inbound */
	pnat=find_inbound(name,iph->protocol,port);
	
	if(pnat){
		/* expired stop DNAT */	
		if(check_expire(pnat)==0){
			pnat->trigger_ip=0;
			return IPT_CONTINUE;
		}
			
		printk("trigger_ip=%u.%u.%u.%u\n",NIPQUAD(pnat->trigger_ip));
#if 1 
		/* Transfer from original range. */
		newrange = ((struct ip_nat_multi_range)
		{ 1, { { IP_NAT_RANGE_MAP_IPS,
			 pnat->trigger_ip, pnat->trigger_ip,
			 {0}, {0} } } });

		/* Hand modified range to generic setup. */
		return ip_nat_setup_info(ct, &newrange, hooknum);
#endif
		return IPT_CONTINUE;	
	}
#if 0
	else if(memcmp("br0",in->name,3)==0){
		/* remeber who triggr this port and save timeout */
		if(trigger_ip==0 || get_expire(ct1)==0){
			ct1=ct;
			trigger_ip=iph->saddr;	
		}	
	}else if(trigger_ip!=0 && get_expire(ct1)!=0){
		

	}
#endif		
	return IPT_CONTINUE;
}

static struct ipt_target pnat_reg
= { { NULL, NULL }, "PNAT", pnat_target, pnat_check, NULL,THIS_MODULE };

static ssize_t proc_read_pnat_fops(struct file *filp,
                                 char *buf,size_t count , loff_t *offp)
{
	char tmp[1024];
	int len=0;
	int i=0;
	
	if(*offp!=0)
		return 0;
	
	for(i=0;i < MAX_PORT_TRIGGER;i++){
		if(pnat_info[i].outbound.min!=0 
				&& pnat_info[i].outbound.max!=0
				&& pnat_info[i].inbound.min!=0
				&& pnat_info[i].inbound.max!=0){

			len+=sprintf(tmp+len,"%s %s[%d-%d] %s %s[%d-%d]\n"
				,pnat_info[i].outbound.ifname
				,pnat_info[i].outbound.protocol
				,pnat_info[i].outbound.min
				,pnat_info[i].outbound.max
				,pnat_info[i].inbound.ifname
				,pnat_info[i].inbound.protocol
				,pnat_info[i].inbound.min
				,pnat_info[i].inbound.max);
		}
	}
	
	copy_to_user(buf,tmp,len);
	*offp = len;
	return len;
}

static ssize_t proc_write_pnat_fops(struct file *filp,const char *buffer,
                                     size_t count , loff_t *offp)
{
	char line[128];
	int i=0;
	
	
	copy_from_user(line,buffer,count);
	line[count]=0x00;
	
	for(i=0;i < MAX_PORT_TRIGGER;i++){
		if(pnat_info[i].outbound.min==0 
				|| pnat_info[i].outbound.max==0
				|| pnat_info[i].inbound.min==0
				|| pnat_info[i].inbound.max==0)
			break;
		/* clean table */
		if(line[0]=='0')
			free_ct_list(&pnat_info[i]);
	}
	
	/* clean table */
	if(line[0]=='0'){
		memset(&pnat_info , 0 , sizeof(struct ipt_pnat_info) * MAX_PORT_TRIGGER);
		printk("Clean\n");
		return count;
	}
		
	
		
	if(i< MAX_PORT_TRIGGER){
		sscanf(line,"%s %s %d-%d %s %s %d-%d"
				,pnat_info[i].outbound.ifname
				,pnat_info[i].outbound.protocol
				,(int *)&pnat_info[i].outbound.min
				,(int *)&pnat_info[i].outbound.max
				,pnat_info[i].inbound.ifname
				,pnat_info[i].inbound.protocol
				,(int *)&pnat_info[i].inbound.min
				,(int *)&pnat_info[i].inbound.max);
		
		if(pnat_info[i].outbound.min==0 
				|| pnat_info[i].outbound.max==0
				|| pnat_info[i].inbound.min==0
				|| pnat_info[i].inbound.max==0){
			printk("Invaild format\n");
			printk("lan_if proto min port-max port  wan_if proto min port-max port\n");
		}
		
	}else
		printk("Only support %d entry\n",MAX_PORT_TRIGGER);

	return count;
};	

static struct file_operations pnat_fops = {
        read: proc_read_pnat_fops,
        write: proc_write_pnat_fops,
};

static struct proc_dir_entry *proc_special;
static struct proc_dir_entry *proc_pnat;

static int __init init(void)
{
	
	proc_special=create_proc_entry("special", S_IFDIR, &proc_root);
	
	if (proc_special == NULL) {
		printk("create_proc_entry() failed\n");
		return -1;
        }
	proc_pnat=create_proc_entry("pnat",0666,proc_special);
	proc_pnat->owner = THIS_MODULE;
	proc_pnat->proc_fops = &pnat_fops;
	
	/* init pnat_info */	
	memset(&pnat_info , 0 , sizeof(struct ipt_pnat_info) * MAX_PORT_TRIGGER);
	
	return ipt_register_target(&pnat_reg);
}

static void __exit fini(void)
{
	int i=0;
	
	remove_proc_entry("pnat",proc_special);
	remove_proc_entry("special",&proc_root);
	
	/* free table */
	for(i=0;i < MAX_PORT_TRIGGER;i++){
		free_ct_list(&pnat_info[i]);
	}

	ipt_unregister_target(&pnat_reg);
}

module_init(init);
module_exit(fini);
MODULE_LICENSE("GPL");
