# asapctl.sh - shell script to control ASAP modules
#
# Copyright (c) 2005, TeamF1, Inc.
#
# modification history
# -------------------
# 01c,20oct05,pha added support for 2.4.x kernel modules
# 01b,08jun05,sud added asapStop
# 01a,23may05,pha written. 
#
#
# asapctl.sh
# ------------
# Script to load ASAP modules on the target.
#

checkKernVer()
{
    if [ -n "`uname -r | grep "2.6"`" ]; then
        MOD_EXT=ko
    elif [ -n "`uname -r | grep "2.4"`" ]; then
        MOD_EXT=o
    else
        echo "Only 2.6.x and 2.4.x kernel versions are supported"
        return 0;
    fi
    UMI_MODNAME=umi.$MOD_EXT
    KLITE_MODNAME=klite.$MOD_EXT
    DOT11_MODNAME=kdot11.$MOD_EXT
    XCALIBUR_MODNAME=xcalibur.$MOD_EXT
    MIBCTL_MODNAME=mibctl.$MOD_EXT
    WLAN_MOD=wlan.$MOD_EXT
    ATH_HAL_MOD=ath_hal.$MOD_EXT
    ATH_RATE_ONOE_MOD=ath_rate_onoe.$MOD_EXT
    ATH_PCI_MOD=ath_pci.$MOD_EXT
    WLAN_XAUTH_MOD=wlan_xauth.$MOD_EXT
    WLAN_CCMP_MOD=wlan_ccmp.$MOD_EXT
    WLAN_TKIP_MOD=wlan_tkip.$MOD_EXT
    WLAN_WEP_MOD=wlan_wep.$MOD_EXT
    WLAN_ACL_MOD=wlan_acl.$MOD_EXT
    return 1;
}

usage ()
{
   echo ""
   echo "Usage:"
   echo -n "sh $0 [-v |--version ] | [ -h | --help ] | "
   echo "[ start | stop ]"
   echo ""
}

version ()
{
    echo "ASAP 1.0.1"
    echo "Copyright (c) 2005, TeamF1, Inc."
    echo "All rights reserved."
    echo "www.TeamF1.com.";   
}


checkModule()
{
    MOD=`lsmod | awk '{print $1".ko"}' | grep "$1"`
    if [ -z "$MOD" ]
    then 
        return 0;
    else
        echo "$1 module already inserted in kernel"
        return 1;
    fi
}

findModule()
{
    MOD=`lsmod | awk '{print $1".ko"}' | grep "$1"`
    if [ -z "$MOD" ]
    then 
        echo "$1 : no such module in the kernel"
        return 1;
    else
        
        return 0;
    fi
}  

loadModule ()
{
    checkModule $1;
    if [ $? -eq 0 ]; then
        if [ ! -f $TGT_MODULE_DIR/$1 ]; then
            echo "$TGT_MODULE_DIR/$1 does not exist"
            return 0;
        else
            insmod $TGT_MODULE_DIR/$1 2>/dev/null
            [ $? -eq 0 ] || { echo "Could not load $1"; asapStartExit 0;}
            echo "$1 loaded into the kernel"
            return 1;
        fi
    else
        return 1;
    fi
}


unLoadModule ()
{
    findModule $1;
    if [ $? -eq 0 ]; then
        rmmod $1 2>/dev/null
        [ $? -eq 0 ] || { echo "Could not unload $1"; asapStopExit 0;}
        echo "$1 unloaded from the kernel"
    fi
    return 1;
}

asapStartExit ()
{
    if [ $1 -eq 0 ]; then
        echo "Failed to start ASAP !!"
        exit 0;
    else
        echo "ASAP started succesfully !"
        exit 0;
    fi
}

asapStopExit ()
{
    if [ $1 -eq 0 ]; then
        echo " this module may be used by other modules"
        exit 0;
    else
        echo "ASAP stopped succesfully !"
        exit 0;
    fi
}

checkEnv()
{
    if [ -z "$2" ]; then
        echo "$1 not set"
        echo "Please refer ASAP Getting Started Guide "
        return 0;
    fi

    return 1;
}

startASAP ()
{
    # Checking env variable TGT_MODULE_DIR
    checkEnv TGT_MODULE_DIR $TGT_MODULE_DIR;
    if [ $? -eq 0 ]; then
        asapStartExit 0;
    fi

    # Checking env variable TGT_SCRIPTS
    checkEnv TGT_SCRIPTS $TGT_SCRIPTS;
    if [ $? -eq 0 ]; then
        asapStartExit 0;
    fi

    # Checking env variable TGT_APPS
    checkEnv TGT_APPS $TGT_APPS;
    if [ $? -eq 0 ]; then
        asapStartExit 0;
    fi

    # Checking env variable TGT_SAMPLES
    checkEnv TGT_SAMPLES $TGT_SAMPLES;
    if [ $? -eq 0 ]; then
        asapStartExit 0;
    fi

    # Checking kernel version
    checkKernVer;
    if [ $? -eq 0 ]; then
        asapStartExit 0;
    fi

    # Loading umi module
    loadModule  $UMI_MODNAME;
    if [ $? -eq 0 ]; then
        asapStartExit 0;
    fi
    sleep 1

    # Loading klite module
    loadModule  $KLITE_MODNAME;
    if [ $? -eq 0 ]; then
        asapStartExit 0;
    fi
    sleep 1

    # Loading dot11 module
    loadModule  $DOT11_MODNAME;
    if [ $? -eq 0 ]; then
        asapStartExit 0;
    fi
    sleep 1

    # Loading xcalibur module
    loadModule  $XCALIBUR_MODNAME;
    if [ $? -eq 0 ]; then
        asapStartExit 0;
    fi
    sleep 1

    # Loading mibctl module
    loadModule  $MIBCTL_MODNAME;
    if [ $? -eq 0 ]; then
        asapStartExit 0;
    fi
    sleep 1

    # Loading wlan module
    loadModule  $WLAN_MOD;
    if [ $? -eq 0 ]; then
        asapStartExit 0;
    fi
    sleep 1

    # Loading ath_hal module
    loadModule  $ATH_HAL_MOD;
    if [ $? -eq 0 ]; then
        asapStartExit 0;
    fi
    sleep 1

    # Loading ath_rate_onoe module
    loadModule  $ATH_RATE_ONOE_MOD;
    if [ $? -eq 0 ]; then
        asapStartExit 0;
    fi
    sleep 1


    # Loading ath_pci module
    loadModule  $ATH_PCI_MOD;
    if [ $? -eq 0 ]; then
        asapStartExit 0;
    fi
    sleep 1


    # Loading wlan_xauth module
    loadModule  $WLAN_XAUTH_MOD;
    if [ $? -eq 0 ]; then
        asapStartExit 0;
    fi
    sleep 1


    # Loading wlan_ccmp module
    loadModule  $WLAN_CCMP_MOD;
    if [ $? -eq 0 ]; then
        asapStartExit 0;
    fi
    sleep 1


    # Loading wlan_tkip module
    loadModule  $WLAN_TKIP_MOD;
    if [ $? -eq 0 ]; then
        asapStartExit 0;
    fi
    sleep 1

    # Loading wlan_wep module
    loadModule  $WLAN_WEP_MOD;
    if [ $? -eq 0 ]; then
        asapStartExit 0;
    fi
    sleep 1

    # Loading wlan_acl module
    loadModule  $WLAN_ACL_MOD;
    if [ $? -eq 0 ]; then
        asapStartExit 0;
    fi
    sleep 1
    
    
    asapStartExit 1;
}

stopASAP ()
{

    # Checking kernel version
    checkKernVer;
    if [ $? -eq 0 ]; then
        exit 0;
    fi

    # unLoading wlan_acl module
    unLoadModule $WLAN_ACL_MOD;
    if [ $? -eq 0 ]; then
        asapStopExit 0;
    fi
    sleep 1

    # unLoading wlan_tkip module
    unLoadModule  $WLAN_TKIP_MOD;
    if [ $? -eq 0 ]; then
        asapStopExit 0;
    fi
    sleep 1

    # unLoading wlan_ccmp module
    unLoadModule  $WLAN_CCMP_MOD;
    if [ $? -eq 0 ]; then
        asapStopExit 0;
    fi
    sleep 1

    # unLoading wlan_xauth module
    unLoadModule  $WLAN_XAUTH_MOD;
    if [ $? -eq 0 ]; then
        asapStopExit 0;
    fi
    sleep 1

    #unLoading wlan_wep module
    unLoadModule  $WLAN_WEP_MOD;
    if [ $? -eq 0 ]; then
        asapStopExit 0;
    fi
    sleep 1

    # unLoading ath_pci module
    unLoadModule  $ATH_PCI_MOD;
    if [ $? -eq 0 ]; then
        asapStopExit 0;
    fi
    sleep 1


    # unLoading ath_rate_onoe module
    unLoadModule  $ATH_RATE_ONOE_MOD;
    if [ $? -eq 0 ]; then
        asapStopExit 0;
    fi
    sleep 1


    # unLoading ath_hal module
    unLoadModule  $ATH_HAL_MOD;
    if [ $? -eq 0 ]; then
        asapStopExit 0;
    fi
    sleep 1

    # unLoading wlan module
    unLoadModule  $WLAN_MOD;
    if [ $? -eq 0 ]; then
        asapStopExit 0;
    fi
    sleep 1
    
    # unLoading xcalibur module
    unLoadModule  $XCALIBUR_MODNAME;
    if [ $? -eq 0 ]; then
        asapStopExit 0;
    fi
    sleep 1


    # unLoading dot11 module
    unLoadModule  $DOT11_MODNAME;
    if [ $? -eq 0 ]; then
        asapStopExit 0;
    fi
    sleep 1

    # unLoading klite module
    unLoadModule  $KLITE_MODNAME;
    if [ $? -eq 0 ]; then
        asapStopExit 0;
    fi
    sleep 1
 
   
    # unLoading umi module
    unLoadModule  $MIBCTL_MODNAME;
    if [ $? -eq 0 ]; then
        asapStopExit 0;
    fi
    sleep 1

    # unLoading umi module
    unLoadModule  $UMI_MODNAME;
    if [ $? -eq 0 ]; then
        asapStopExit 0;
    fi
    sleep 1

    asapStopExit 1;
}



#
#  Main Function 
#

case "$1" in

    -h | --help ) usage; exit 0;;

    -v | --version ) version; exit 0;;
    
    start) startASAP;;

    stop) stopASAP;;
    
       *) usage; ;;

esac
